#include "diskimage.h"

#include "zlib.h"

#include <QFileInfo>
#include "aspeqtsettings.h"
#include <QDir>

/* Compression - uncompression functions */

bool DiskImage::compressGzip (const QString &in, const QString &out)
{
    QFile outFile(out);
    if (!outFile.open(QFile::WriteOnly | QFile::Truncate)) {
        emitUiMessage(tr("Cannot open image file '%1' to write back: %2")
                       .arg(out)
                       .arg(outFile.errorString()), MessageType::ErrorMessage);
        return false;
    }

    gzFile zipped = gzdopen(outFile.handle(), "wb9");
    if (!zipped) {
        emitUiMessage(tr("Cannot save image file '%1' back: %2.").arg(out), MessageType::ErrorMessage);
        return false;
    }
    QByteArray buffer;

    QFile inFile(in);

    if (!inFile.open(QFile::ReadOnly)) {
        emitUiMessage(tr("Cannot save image file '%1' back: Cannot open temporary file '%2': %3")
                       .arg(out)
                       .arg(in)
                       .arg(inFile.errorString()), MessageType::ErrorMessage);
        gzclose(zipped);
        return false;
    }

    int r;
    do {
        buffer = inFile.read(65536);
        if (buffer.size() < 65536 && !inFile.atEnd()) {
            emitUiMessage(tr("Cannot save image file '%1' back: Cannot read from temporary file '%2': %3")
                           .arg(out)
                           .arg(in)
                           .arg(inFile.errorString()), MessageType::ErrorMessage);
            gzclose(zipped);
            return false;
        }
        r = gzwrite(zipped, buffer.data(), buffer.size());
        if (r != buffer.size()) {
            emitUiMessage(tr("Cannot write back to image file '%1'")
                           .arg(out), MessageType::ErrorMessage);
            gzclose(zipped);
            return false;
        }
    } while (r == 65536);

    gzclose(zipped);
    return true;
}

bool DiskImage::uncompressGzip (const QString &in, const QString &out)
{
    QFile inFile(in);
    if (!inFile.open(QFile::ReadOnly)) {
        emitUiMessage(tr("Cannot open image file '%1': %2")
                       .arg(in)
                       .arg(inFile.errorString()), MessageType::ErrorMessage);
        return false;
    }
    int fd = inFile.handle();
    gzFile zipped = gzdopen(fd, "rb");
    if (!zipped) {
        emitUiMessage(tr("Cannot open image file '%1': Decompression failed.").arg(in), MessageType::ErrorMessage);
        return false;
    }
    QByteArray buffer;
    buffer.resize(65536);

    QFile outFile(out);

    if (!outFile.open(QFile::WriteOnly | QFile::Truncate)) {
        emitUiMessage(tr("Cannot open image file '%1': Cannot open temporary image file '%2': %3")
                       .arg(in)
                       .arg(out)
                       .arg(outFile.errorString()), MessageType::ErrorMessage);
        gzclose(zipped);
        return false;
    }

    int r;
    do {
        r = gzread(zipped, buffer.data(), 65536);
        if (r == -1) {
            emitUiMessage(tr("Cannot open image file '%1': Decompression failed.").arg(in), MessageType::ErrorMessage);
            gzclose(zipped);
            outFile.close();
            outFile.remove();
            return false;
        }
        if (outFile.write(buffer.constData(), r) != r) {
            emitUiMessage(tr("Cannot open image file '%1': Cannot write to temporary image file '%2': %3")
                           .arg(in)
                           .arg(out)
                           .arg(outFile.errorString()), MessageType::ErrorMessage);
            gzclose(zipped);
            outFile.close();
            outFile.remove();
            return false;
        }
    } while (r == 65536);

    gzclose(zipped);
    return true;
}

bool DiskImage::compressDcm (const QString &in, const QString &out)
{
    return false;
}

bool DiskImage::uncompressDcm (const QString &in, const QString &out)
{
    return false;
}

bool DiskImage::compressScp (const QString &in, const QString &out)
{
    return false;
}

bool DiskImage::uncompressScp (const QString &in, const QString &out)
{
    return false;
}

/* AtrImage */

bool AtrImage::format(quint16 aSectorCount, quint16 aSectorSize)
{
    if (!RawDiskImage::format(aSectorCount, aSectorSize)) {
        return false;
    }
    int size = file.size();
    if (!file.resize(size + 16)) {
        emitUiMessage(tr("Cannot resize %1: %2")
                       .arg(file.fileName(), file.errorString()), MessageType::ErrorMessage);
        return false;
    }
    if (!file.seek(0)) {
        emitUiMessage(tr("Cannot rewind %1: %2")
                       .arg(file.fileName(), file.errorString()), MessageType::ErrorMessage);
        return false;
    }
    QByteArray header(16, 0);
    size = size / 16;
    header[0] = 0x96;
    header[1] = 0x02;
    header[2] = size % 256;
    header[3] = (size / 256) % 256;
    header[4] = aSectorSize % 256;
    header[5] = aSectorSize / 256;
    header[6] = (size / 65536) % 256;
    header[7] = size / 16777216;
    if (file.write(header) != 16) {
        emitUiMessage(tr("Cannot write header to %1: %2")
                       .arg(file.fileName(), file.errorString()), MessageType::ErrorMessage);
        return false;
    }
    mSectorSize = aSectorSize;
    return true;
}

quint16 AtrImage::sectorCount()
{
    uint rawSize = file.size() - 16;

    if (rawSize <= 128 * 3 || sectorSize() == 128) {
        return (rawSize + 127) / 128;
    }

    return ((rawSize - 128 * 3) + sectorSize() - 1) / sectorSize() + 3;
}

quint16 AtrImage::sectorSize()
{
    return mSectorSize;
}

bool AtrImage::seekToSector(quint16 sector)
{
    uint offset = (sector - 1) * 128 + 16;

    if (sector > 4 && sectorCount() != 128) {
        offset += (sector - 4) * (sectorSize() - 128);
    }

    if (!file.seek(offset)) {
        emitUiMessage(tr("Cannot seek to sector %1 in '%2': %3").arg(QString().number(sector), file.fileName(), file.errorString()), MessageType::ErrorMessage);
        return false;
    }

    return true;
}

bool AtrImage::open(const QString &fileName, CompressionType compression, bool readOnly)
{
    if (!RawDiskImage::open(fileName, compression, readOnly)) {
        return false;
    }

    /* Read header */
    QByteArray header;
    header = file.read(16);
    if (header.size() != 16) {
        emitUiMessage(tr("Cannot read the header of '%1'.").arg(file.fileName()), MessageType::ErrorMessage);
        close();
        return false;
    }

    quint16 magic = ((quint8)header[0]) + ((quint8)header[1]) * 256;

    if (magic != 0x0296) {
        emitUiMessage(tr("'%1' is not a valid ATR file: Magic number does not match.").arg(file.fileName()), MessageType::ErrorMessage);
        close();
        return false;
    }

    quint16 sizeLo = ((quint8)header[2]) + ((quint8)header[3]) * 256;
    quint16 secSize = ((quint8)header[4]) + ((quint8)header[5]) * 256;
    quint32 sizeHi = ((quint8)header[6]) + ((quint8)header[7]) * 256;
    quint32 size = sizeLo + sizeHi * 65536;

    if (size * 16 != file.size() - 16) {
        emitUiMessage(tr("Size of '%1' is reported as %2 bytes in the header but it's actually %3.")
                       .arg(file.fileName())
                       .arg(size * 16)
                       .arg(file.size() - 16), MessageType::WarningMessage);
    }

    mSectorSize = secSize;

    return true;
}

/* DiskImage */

void DiskImage::handleCommand(quint8 command, quint16 aux)
{
    switch (command) {
        case 0x22:
            {   /* Format ED */
                if (!sio->port()->writeCommandAck()) {
                    return;
                }
                if (isReadOnly()) {
                    sio->port()->writeError();
                    emitUiMessage(QString("[%1] Format ED denied.")
                                   .arg(deviceName()), MessageType::WarningMessage);
                    break;
                }
                if (format(1040, 128)) {
                    QByteArray data(newSectorSize, 0);
                    data[0] = data[1] = 0xFF;
                    sio->port()->writeComplete();
                    sio->port()->writeDataFrame(data);
                    emitUiMessage(QString("[%1] Format 1040 sectors * 128 bytes.")
                                   .arg(deviceName()), MessageType::NormalMessage);
                } else {
                    sio->port()->writeError();
                    emitUiMessage(QString("[%1] Format 1040 sectors * 128 bytes failed.")
                                   .arg(deviceName()), MessageType::ErrorMessage);
                }
                break;
            }
        case 0x3F:
            {    /* Get high speed index */
                if (!sio->port()->writeCommandAck()) {
                    return;
                }
                sio->port()->writeComplete();
                QByteArray speed(1, 0);
                switch (aspeqtSettings->maximumSerialPortSpeed()) {
                    case 0:
                        speed[0] = 0x28;    // 19200
                        break;
                    case 1:
                        speed[0] = 0x10;    // 38400
                        break;
                    case 2:
                    default:
                        speed[0] = 0x08;    // 57600
                        break;
                }
                sio->port()->writeDataFrame(speed);
                emitUiMessage(QString("[%1] Speed poll.").arg(deviceName()), MessageType::NormalMessage);
                break;
            }
        case 0x4e:
            {    /* Get configuration */
                if (!sio->port()->writeCommandAck()) {
                    return;
                }

                int sideCount;
                int tracksPerSide;
                int sectorsPerTrack;
                int secs = sectorCount();
                int bytes = sectorSize();

                if (newSectorCount) {
                    secs = newSectorCount;
                    bytes = newSectorSize;
                }

                switch (secs) {
                    case 630:
                        // 35*18*1
                        sideCount = 1;
                        tracksPerSide = 35;
                        sectorsPerTrack = 18;
                        break;
                    case 720:
                        // 40*18*1
                        sideCount = 1;
                        tracksPerSide = 40;
                        sectorsPerTrack = 18;
                        break;
                    case 910:
                        // 35*26*1
                        sideCount = 1;
                        tracksPerSide = 35;
                        sectorsPerTrack = 26;
                        break;
                    case 1040:
                        //  40*26*1
                        sideCount = 1;
                        tracksPerSide =  40;
                        sectorsPerTrack = 26;
                        break;
                    case 1260:
                        //  35*18*2
                        sideCount = 2;
                        tracksPerSide =  30;
                        sectorsPerTrack = 18;
                        break;
                    case 1440:
                        //  40*18*2
                        sideCount = 2;
                        tracksPerSide =  40;
                        sectorsPerTrack = 18;
                        break;
                    case 1820:
                        //  35*26*2
                        sideCount = 2;
                        tracksPerSide =  35;
                        sectorsPerTrack = 26;
                        break;
                    case 2002:
                        //  77*26*1
                        sideCount = 1;
                        tracksPerSide =  77;
                        sectorsPerTrack = 26;
                        break;
                    case 2080:
                        //  40*26*2
                        sideCount = 2;
                        tracksPerSide =  40;
                        sectorsPerTrack = 26;
                        break;
                    case 2880:
                        //  80*18*2
                        sideCount = 2;
                        tracksPerSide =  80;
                        sectorsPerTrack = 18;
                        break;
                    case 4004:
                        //  77*26*2
                        sideCount = 2;
                        tracksPerSide =  77;
                        sectorsPerTrack = 26;
                        break;
                    case 4160:
                        //  80*26*2
                        sideCount = 2;
                        tracksPerSide =  80;
                        sectorsPerTrack = 26;
                        break;
                    default:
                        sideCount = 1;
                        tracksPerSide = 1;
                        sectorsPerTrack = sectorCount();
                }

                QByteArray data(12, 0);
                data[0] = tracksPerSide;
                data[1] = 1;    // Step rate
                data[2] = sectorsPerTrack / 256;
                data[3] = sectorsPerTrack % 256;
                data[4] = sideCount - 1;
                data[5] = (bytes == 256) * 4 | (tracksPerSide == 77) * 2;
                data[6] = bytes / 256;
                data[7] = bytes % 256;
                data[8] = 0xff;
                sio->port()->writeComplete();
                sio->port()->writeDataFrame(data);
                emitUiMessage(QString("[%1] Get configuration.")
                               .arg(deviceName()), MessageType::NormalMessage);
                break;
            }
        case 0x4f:
        case 0x66:
            {    /* Set configuration */
                int size;
                if (command == 0x4f) {
                    size = 12;
                } else {
                    size = 128;
                }
                if (!sio->port()->writeCommandAck()) {
                    return;
                }
                QByteArray data = sio->port()->readDataFrame(size);
                if (data.isEmpty()) {
                    sio->port()->writeDataNak();
                    break;
                }
                sio->port()->writeDataAck();
                int sideCount = (quint8)data[4] + 1;
                int tracksPerSide = (quint8)data[0];
                int sectorsPerTrack = (quint8)data[3] + 256 * (quint8)data[2];
                newSectorCount = sideCount * tracksPerSide * sectorsPerTrack;
                newSectorSize = (quint8)data[7] + 256*(quint8)data[6];
                if (newSectorCount < 65536 && (newSectorSize == 128 || newSectorSize == 256)) {
                    if (command == 0x4f) {
                        sio->port()->writeComplete();
                        emitUiMessage(QString("[%1] Set configuration to %2 sectors * %3 bytes.")
                                       .arg(deviceName())
                                       .arg(newSectorCount)
                                       .arg(newSectorSize), MessageType::NormalMessage);
                    } else {
                        if (format(newSectorCount, newSectorSize)) {
                            sio->port()->writeComplete();
                            emitUiMessage(QString("[%1] Skewed format %2 sectors * %3 bytes.")
                                           .arg(deviceName())
                                           .arg(newSectorCount)
                                           .arg(newSectorSize), MessageType::NormalMessage);
                        } else {
                            sio->port()->writeError();
                            emitUiMessage(QString("[%1] Skewed format %2 sectors * %3 bytes failed.")
                                           .arg(deviceName())
                                           .arg(newSectorCount)
                                           .arg(newSectorSize), MessageType::ErrorMessage);
                        }
                    }
                } else {
                    sio->port()->writeError();
                    emitUiMessage(QString("[%1] Set configuration to %2 sectors * %3 bytes failed.")
                                   .arg(deviceName())
                                   .arg(newSectorCount)
                                   .arg(newSectorSize), MessageType::WarningMessage);
                    break;
                }
                break;
            }
        case 0x21:
            {   /* Format */
                if (!sio->port()->writeCommandAck()) {
                    return;
                }
                if (isReadOnly()) {
                    sio->port()->writeError();
                    emitUiMessage(QString("[%1] Format denied.")
                                   .arg(deviceName()), MessageType::WarningMessage);
                    break;
                }
                if (newSectorCount == 0) {
                    newSectorCount = 720;
                    newSectorSize = 128;
                }
                if (format(newSectorCount, newSectorSize)) {
                    QByteArray data(newSectorSize, 0);
                    data[0] = data[1] = 0xFF;
                    sio->port()->writeComplete();
                    sio->port()->writeDataFrame(data);
                    emitUiMessage(QString("[%1] Format %2 sectors * %3 bytes.")
                                   .arg(deviceName())
                                   .arg(newSectorCount)
                                   .arg(newSectorSize), MessageType::NormalMessage);
                } else {
                    sio->port()->writeError();
                    emitUiMessage(QString("[%1] Format %2 sectors * %3 bytes failed.")
                                   .arg(deviceName())
                                   .arg(newSectorCount)
                                   .arg(newSectorSize), MessageType::ErrorMessage);
                }
                break;
            }
        case 0x50:
        case 0x57:
            {    /* Write sector */
                if (aux >= 1 && aux <= sectorCount()) {
                    if (!sio->port()->writeCommandAck()) {
                        return;
                    }
                    int secSize = 128;
                    if (aux > 3) {
                        secSize = sectorSize();
                    }
                    QByteArray data = sio->port()->readDataFrame(secSize);
                    if (!data.isEmpty()) {
                        sio->port()->writeDataAck();
                        if (isReadOnly()) {
                            sio->port()->writeError();
                            emitUiMessage(QString("[%1] Write sector %2 denied.")
                                           .arg(deviceName())
                                           .arg(aux), MessageType::WarningMessage);
                            break;
                        }
                        if (writeSector(aux, data)) {
                            sio->port()->writeComplete();
                            emitUiMessage(QString("[%1] Write sector %2 (%3 bytes).")
                                           .arg(deviceName())
                                           .arg(aux)
                                           .arg(data.size()), MessageType::NormalMessage);
                        } else {
                            sio->port()->writeError();
                            emitUiMessage(QString("[%1] Write sector %2 failed.")
                                           .arg(deviceName())
                                           .arg(aux), MessageType::ErrorMessage);
                        }
                    } else {
                        emitUiMessage(QString("[%1] Write sector %2 data frame failed.")
                                       .arg(deviceName())
                                       .arg(aux), MessageType::ErrorMessage);
                        sio->port()->writeDataNak();
                    }
                } else {
                    sio->port()->writeCommandNak();
                    emitUiMessage(QString("[%1] Write sector %2 NAKed.")
                                   .arg(deviceName())
                                   .arg(aux), MessageType::WarningMessage);
                }
                break;
            }
        case 0x52:
            {    /* Read sector */
                if (aux >= 1 && aux <= sectorCount()) {
                    if (!sio->port()->writeCommandAck()) {
                        return;
                    }
                    QByteArray data;
                    if (readSector(aux, data)) {
                        sio->port()->writeComplete();
                        sio->port()->writeDataFrame(data);
                        emitUiMessage(QString("[%1] Read sector %2 (%3 bytes).")
                                       .arg(deviceName())
                                       .arg(aux)
                                       .arg(data.size()), MessageType::NormalMessage);
                    } else {
                        emitUiMessage(QString("[%1] Read sector %2 failed.")
                                       .arg(deviceName())
                                       .arg(aux), MessageType::ErrorMessage);
                        sio->port()->writeError();
                    }
                } else {
                    sio->port()->writeCommandNak();
                    emitUiMessage(QString("[%1] Read sector %2 NAKed.")
                                   .arg(deviceName())
                                   .arg(aux), MessageType::WarningMessage);
                }
                break;
            }
        case 0x53:
            {    /* Get status */
                if (!sio->port()->writeCommandAck()) {
                    return;
                }

                int secs = sectorCount();
                int bytes = sectorSize();

                if (newSectorCount) {
                    secs = newSectorCount;
                    bytes = newSectorSize;
                }

                QByteArray status(4, 0);
                status[0] = isReadOnly() * 8 |
                            (bytes == 256) * 32 |
                            (bytes == 128 && secs == 1040) * 128;
                status[3] = 1;
                sio->port()->writeComplete();
                sio->port()->writeDataFrame(status);
                emitUiMessage(QString("[%1] Get status.")
                               .arg(deviceName()), MessageType::NormalMessage);
                break;
            }
        default:        /* Unknown command */
            sio->port()->writeCommandNak();
            emitUiMessage(QString("[%1] command: $%2, aux: $%3 NAKed.")
                           .arg(deviceName())
                           .arg(command, 2, 16, QChar('0'))
                           .arg(aux, 4, 16, QChar('0')), MessageType::WarningMessage);
            break;
    }
}

/* RawDiskImage */

RawDiskImage::~RawDiskImage()
{
    if (isOpen()) {
        close();
    }
}

bool RawDiskImage::close()
{
    file.close();
    if (tempFile.isOpen()) {
        switch (compressionType) {
            case GzipCompression:
                if (isDirty && !compressGzip(tempFile.fileName(), originalName)) {
                    tempFile.setAutoRemove(false);
                    tempFile.close();
                    return false;
                }
                break;
            case ScpCompression:
                if (isDirty && !compressScp(tempFile.fileName(), originalName)) {
                    tempFile.setAutoRemove(false);
                    tempFile.close();
                    return false;
                }
                break;
            case DcmCompression:
                if (isDirty && !compressDcm(tempFile.fileName(), originalName)) {
                    tempFile.setAutoRemove(false);
                    tempFile.close();
                    return false;
                }
                break;
            default:
                tempFile.setAutoRemove(false);
                tempFile.close();
                return false;
        }
        tempFile.close();
    }

    return true;
}

bool RawDiskImage::create(const QString &fileName, quint16 aSectorCount, quint16 aSectorSize, CompressionType compression)
{
    if (isOpen()) {
        close();
    }

    compressionType = GzipCompression;

    if (compression == NoCompression) {
        file.setFileName(fileName);
    } else {
        tempFile.setFileTemplate(QDir::temp().absoluteFilePath("aspeqt-temp-XXXXXX"));
        tempFile.open();
        originalName = fileName;
        file.setFileName(tempFile.fileName());
    }

    if (!file.open(QFile::ReadWrite)) {
        emitUiMessage(tr("Cannot open '%1': %2")
                       .arg(file.fileName())
                       .arg(file.errorString()), MessageType::ErrorMessage);
        return false;
    }
    if (!format(aSectorCount, aSectorSize)) {
        return false;
    }
    close();
    if (!open(fileName, compression, false)) {
        return false;
    }
    return true;
}

bool RawDiskImage::format(quint16 aSectorCount, quint16 aSectorSize)
{
    int size = aSectorCount * aSectorSize;
    if (aSectorSize != 128) {
        int dif = 3;
        if (aSectorCount < 3) {
            dif = aSectorCount;
        }
        size -= dif * (aSectorSize - 128);
    }
    if (file.resize(0) && file.resize(size)) {
        isDirty = true;
        emit statusChanged(mDeviceNo);
        return true;
    } else {
        emitUiMessage(tr("Cannot resize %1: %2")
                       .arg(file.fileName(), file.errorString()), MessageType::ErrorMessage);
        return false;
    }
}

bool RawDiskImage::isOpen()
{
    return file.isOpen();
}

bool RawDiskImage::isReadOnly()
{
    return !file.isWritable();
}

bool RawDiskImage::open(const QString &fileName, CompressionType compression, bool readOnly)
{
    if (isOpen()) {
        close();
    }

    QFileInfo info(fileName);
    QFile::OpenMode mode;

    if (!info.exists()) {
        emitUiMessage(tr("Cannot open image file '%1': File does not exist.").arg(fileName), MessageType::ErrorMessage);
        return false;
    }

    compressionType = compression;

    switch (compressionType) {
        case GzipCompression:
            tempFile.setFileTemplate(QDir::temp().absoluteFilePath("aspeqt-temp-XXXXXX"));
            tempFile.open();
            if (!uncompressGzip(fileName, tempFile.fileName())) {
                tempFile.close();
                return false;
            }
            break;
        case ScpCompression:
            break;
        case DcmCompression:
            break;
        default:
            break;
    }

    if (info.isWritable() || readOnly) {
        mode = QFile::ReadWrite;
    } else {
        mode = QFile::ReadOnly;
    }

    if (compressionType == NoCompression) {
        file.setFileName(fileName);
    } else {
        file.setFileName(tempFile.fileName());
        originalName = fileName;
    }

    if (!file.open(mode)) {
        emitUiMessage(tr("Cannot open image file '%1': %2").arg(fileName, file.errorString()), MessageType::ErrorMessage);
        return false;
    }

    newSectorCount = 0;
    newSectorSize = 0;

    isDirty = false;

    emit statusChanged(mDeviceNo);

    return true;
}

QString RawDiskImage::properties()
{
    QString comp;
    switch (compressionType) {
        case NoCompression:
            comp = imageType();
            break;
        case GzipCompression:
            comp = tr("compressed %1").arg(imageType());
            break;
        case DcmCompression:
            comp = "DCM";
            break;
        default:
            break;
    }
    return tr("%1 sectors * %2 bytes %3 disk image")
            .arg(sectorCount())
            .arg(sectorSize())
            .arg(comp);
}

bool RawDiskImage::readSector(quint16 sector, QByteArray &data)
{
    if (!seekToSector(sector)) {
        return false;
    };

    quint16 size = 128;
    if (sector > 3) {
        size = sectorSize();
    }

    quint64 bytesToRead = size;

    if (file.bytesAvailable() < size) {
        bytesToRead = file.bytesAvailable();
    }

    data = file.read(bytesToRead);

    if (data.size() != (int)bytesToRead) {
        emitUiMessage(tr("Cannot read from '%1': %2").arg(file.fileName(), file.errorString()), MessageType::ErrorMessage);
        return false;
    }

    data.resize(size);

    return true;
}

quint16 RawDiskImage::sectorCount()
{
    return (file.size() + 127) / 128;
}

quint16 RawDiskImage::sectorSize()
{
    int result;
    quint64 size = file.size();

    switch (size) {
        case 80640:
        case 92160:
        case 116480:
        case 133120:
        case 232960:
        case 256256:
        case 266240:
        case 512640:
        case 532480:
            result = 128;
            break;
        case 161280:
        case 184320:
        case 322560:
        case 368640:
        case 512512:
        case 737280:
        case 1025024:
            result = 256;
            break;
        default:
            if (size <= 133120) {
                result = 128;
            } else {
                result = 256;
            }
            break;
    }

    return result;
}

bool RawDiskImage::seekToSector(quint16 sector)
{
    uint offset = (sector - 1) * 128;

    if (sector > 3 && sectorCount() != 128) {
        offset += (sector - 3) * (sectorSize() - 128);
    }

    if (!file.seek(offset)) {
        emitUiMessage(tr("Cannot seek to sector %1 in '%2': %3").arg(QString().number(sector), file.fileName(), file.errorString()), MessageType::ErrorMessage);
        return false;
    }

    return true;
}

bool RawDiskImage::setReadOnly(bool readOnly)
{
    file.close();
    bool result;

    if (readOnly) {
        result = file.open(QFile::ReadOnly);
    } else {
        result = file.open(QFile::ReadWrite);
    }
    if (!result) {
        emitUiMessage(tr("Cannot open image file '%1': %2").arg(file.fileName(), file.errorString()), MessageType::ErrorMessage);
    }
    return result;
}

bool RawDiskImage::writeSector(quint16 sector, const QByteArray &data)
{

    if (!seekToSector(sector)) {
        return false;
    };

    quint16 size = 128;
    if (sector > 3 && sectorSize() != 256) {
        size = sectorSize();
    }

    quint64 bytesWritten = file.write(data);

    if (data.size() != (int)bytesWritten) {
        emitUiMessage(tr("Cannot write to '%1': %2").arg(file.fileName(), file.errorString()), MessageType::ErrorMessage);
        return false;
    }

    isDirty = true;

    return true;
}

/* XfdImage */

bool XfdImage::open(const QString &fileName, CompressionType compression, bool readOnly)
{
    if (RawDiskImage::open(fileName, compression, readOnly)) {
        if (file.size() % 128) {
            emitUiMessage(tr("There are partial sectors in '%1'.").arg(file.fileName()), MessageType::ErrorMessage);
        }
        return true;
    } else {
        return false;
    }
}
