#ifndef DISKIMAGE_H
#define DISKIMAGE_H

#include <QFile>
#include <QTemporaryFile>

#include "sioworker.h"

class DiskImage: public SioDevice
{
    Q_OBJECT
    Q_ENUMS(CompressionType)

protected:
    int newSectorSize;
    int newSectorCount;
    bool isDirty;

    bool compressGzip (const QString &in, const QString &out);
    bool uncompressGzip (const QString &in, const QString &out);
    bool compressDcm (const QString &in, const QString &out);
    bool uncompressDcm (const QString &in, const QString &out);
    bool compressScp (const QString &in, const QString &out);
    bool uncompressScp (const QString &in, const QString &out);

public:
    DiskImage(SioWorker *worker): SioDevice(worker) {}

    enum CompressionType {NoCompression, GzipCompression, DcmCompression, ScpCompression};

    virtual bool close() = 0;
    virtual bool format(quint16 aSectorCount, quint16 aSectorSize) = 0;
    void handleCommand(quint8 command, quint16 aux);
    virtual bool isOpen() = 0;
    virtual bool isReadOnly() = 0;
    virtual bool open(const QString &fileName, CompressionType compression, bool readOnly) = 0;
    virtual QString properties() = 0;
    virtual bool readSector(quint16 sector, QByteArray &data) = 0;
    virtual quint16 sectorCount() = 0;
    virtual quint16 sectorSize() = 0;
    virtual bool setReadOnly(bool readOnly) = 0;
    virtual bool writeSector(quint16 sector, const QByteArray &data) = 0;
};

class RawDiskImage: public DiskImage
{
    Q_OBJECT

protected:
    CompressionType compressionType;
    QFile file;
    QTemporaryFile tempFile;
    QString originalName;

    virtual bool seekToSector(quint16 sector);

public:
    RawDiskImage(SioWorker *worker): DiskImage(worker) {}
    ~RawDiskImage();

    bool close();
    bool create(const QString &fileName, quint16 aSectorCount, quint16 aSectorSize, CompressionType compression);
    bool format(quint16 aSectorCount, quint16 aSectorSize);
    virtual QString imageType() {return "raw";}
    bool isOpen();
    bool isReadOnly();
    bool open(const QString &fileName, CompressionType compression, bool readOnly);
    QString properties();
    bool readSector(quint16 sector, QByteArray &data);
    quint16 sectorCount();
    quint16 sectorSize();
    bool setReadOnly(bool readOnly);
    bool writeSector(quint16 sector, const QByteArray &data);
};

class XfdImage: public RawDiskImage
{
    Q_OBJECT

public:
    XfdImage(SioWorker *worker): RawDiskImage(worker) {}

    QString imageType() {return "XFD";}
    bool open(const QString &fileName, CompressionType compression, bool readOnly);
};

class AtrImage: public RawDiskImage
{
    Q_OBJECT

protected:
    uint mSectorSize;

    bool seekToSector(quint16 sector);

public:
    AtrImage(SioWorker *worker): RawDiskImage(worker) {}

    bool format(quint16 aSectorCount, quint16 aSectorSize);
    QString imageType() {return "ATR";}
    quint16 sectorCount();
    quint16 sectorSize();
    bool open(const QString &fileName, CompressionType compression, bool readOnly);
};

#endif // DISKIMAGE_H
